#
# <20230115.1454.46>
#
# An effort to sanely find UNC ports for windows.
# First we test a couple of register values to see if
# smb is supported and if so SMBv1 or SMBv2
# set wh HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\LanmanServer\\Parameters
# registry get $wh smb1
# If we have SMBv1, there is a good chance there is a browser
# If not, we have to test each visable machine on the local network
#
# We use ipconfig, arp, and ping to try and find these with out
# spending forever doing it.

# then we need to check each machine

namespace eval findUNC {
  variable lanmanParms [join {
    HKEY_LOCAL_MACHINE
    SYSTEM
    CurrentControlSet
    Services
    LanmanServer
    Parameters
  } "\\"]
  # nonServers are devices which respond to ping but not to
  # server enumeration. They are alive but have no resources
  # for us.
  variable wslNativeMap {}
  variable nonServers {}

  # proc isSMB {} {
  #   variable lanmanParms
    
  #   # First test if we have SMBv1
  #   set r [catch {registry values $lanmanParms} vals]
  #   # A problem here means we don't have lanman at all...
  #   if {$r != 0} {return 0}
  #   if {[catch {registry get $lanmanParms SMB1} vals] == 0 && $vals == 1} {
  #     # Looks like SMBv1 is alive and well...
  #     return 1
  #   }
  #   if {[catch {registry get $lanmanParms SMB2} vals] == 0 && $vals == 1} {
  #     # No 1, but we have SMB2
  #     return 2
  #   }
  #   # Neither 1 nor 2, likely not 3 as well, but we test anyway
  #   if {[catch {registry get $lanmanParms SMB3} vals] == 0 && $vals == 1} {
  #     # No 1, or 2 but we have SMB3
  #     return 3
  #   }
  #   return 0
  # }

  # This routine checks if an IPv4 'ip' is in the sub net defined by the 'mask'
  proc inSubnet {ip} {
    variable maskParts
    variable myNetParts
    
    foreach ipPart [split $ip .] maskPart $maskParts myNetPart $myNetParts {
      frputs ipPart myNetPart maskPart ip 
      if {(($ipPart ^ $myNetPart) & $maskPart) != 0} {
	frputs ipPart myNetPart maskPart 
	return 0
      }
    }
    frputs "YES " ip
    return 1
  }

  # ipToNetBios takes an ip address and translates it to the NetBios name
  # but only if there is one. Otherwise, it returns nil i.e. {}
  # We use "nbtstat" to do the translation. This, of course, polls the owner
  # so this could take a while, especially if the device is offline or not
  # a NetBios server.
  # Most of the work here is decoding the returned info.
  
  proc ipToNetBios {ip} {
    set raw [asyncExec nbtstat -A $ip]
    if {[regexp {.*\n([^<]*)<20>.*} $raw m name]} {
      return [string trim $name]
    }
    return {}
  }

  # Here we use the registry to find all the local Ip addresses
  # To be usable there must be a gateway for the IP
  
  proc getLocalIps {} {
    # registery interface keys are here:
    
    set ifKey\
	HKEY_LOCAL_MACHINE/SYSTEM/CurrentControlSet/Services/Tcpip/Parameters/Interfaces
    lassign {} IpVals
    foreach iFace [registry keys   [file native $ifKey]] {
      set ifData [registry values  [file native $ifKey/$iFace]]
      foreach gate  [lsearch -glob -all -inline $ifData *DefaultGateway] {
	if {[set val [registry get [file native $ifKey/$iFace] $gate]] != {}} {
	  set prefix [regsub {DefaultGateway} $gate {}]
	  if {"${prefix}IPAddress" in $ifData &&\
		  "${prefix}SubnetMask" in $ifData &&\
		  [set Ip [registry get   [file native $ifKey/$iFace]\
		       "${prefix}IPAddress" ]] != {} &&\
		  [set Mask [registry get [file native $ifKey/$iFace]\
		       "${prefix}SubnetMask" ]] != {}} {		  
	    lappend IpVals [list $Ip $Mask $val]
	  }
	}
      }
    }
    return $IpVals
  }
  
  proc UNCpoll {args} {
    variable nonServers
    variable maskParts {}
    variable myNetParts {}


    # So here we compute the broadcast address...
    # and set up the *Parts variables
    #
    # use twapi  to get the arp table
    set arpList [lindex [twapi::get_arp_table] 1]
    # frputs adapterList
    
    foreach adds [getLocalIps] {
      frputs adds
      lassign $adds myIp mask
      lassign {} bcP myNetParts  maskParts
      #set bcP [set myNetParts [set maskParts {}]]
      foreach myP [split $myIp .] maskP [split $mask .] {
	lappend myNetParts [expr {$myP & $maskP}]
	lappend maskParts $maskP
	lappend bcP [expr {$myP | (~$maskP & 255)}]
      }
      set bc [join $bcP .]
      # ping the broadcast address. One time is enough
      catch {asyncExec ping -n 1 -i 1 -4 $bc}
      # each subList is: interfaceIndex MAC IP type

      # prune the arp entries that are not in the subnet
      # We do this here so we can get a good handle on how many
      # we need to check. Makes the progress bar closer to real
      set newApList {}
      foreach line $arpList {
	if {![inSubnet [set tad [lindex $line 2]]] ||\
		$tad in "$bc $myIp" ||\
		$tad in $nonServers} {continue}
	lappend newApList $line
      }
      #
      # For each of these that are in the subnet see if
      # they have any shares to contribute
      set serverList {}
      set total [llength $newApList]
      set current 0
      foreach line $newApList {
	incr current
	lassign $line du mac tad type
	# it appears that type==invalid means it is stale...
	if {$type == "invalid"} {continue}
	LogStatusOnly "Checking $tad" [expr {($current * 100) / $total}]
	frputs tad bc myIp
	set r [catch {asyncExec ping -a -n 1 -w 1 -4 $tad} pingResult]
	frputs r tab pingResult
	if {$r != 0} {continue}
	# now the time consuming part
	set r [catch {ipToNetBios $tad} nbiosName]
	frputs tad  nbiosName r
	if {$r != 0 || $nbiosName == {}} {
	  lappend nonServers $tad
	  continue
	}
	# OK we have netBIOS name, now get the shares...
	frputs nbiosName
	set raw [asyncExec net view $nbiosName]
	foreach line [split $raw \n] {
	  if {[regexp {(.*)Disk(.*)} $line n share driveletter]} {
	    if {[string index [set trdr [string trim $driveletter]] 1] == ":"} {
	      set driveletter [string index $trdr 0]:
	    } else {
	      set driveletter {}
	    }
	    set share [string trim $share]
	    frputs share driveletter
	    lappend serverList [list //$nbiosName/$share $driveletter]
	  }
	}
      }
    }
    # Lets add the wsl shares
    findWSLshares
    variable wslNativeMap
    foreach share $wslNativeMap {
      lappend serverList [list [lindex $share 0] {}]
    }
    LogStatusOnly "done" -1
    return $serverList
  }

  # we return a DIC (i.e. share  driveletter)  driveletter may be {} if none
  proc getShares {args} {
    variable nonServers
    # if {[isSMB] == 0} {
    #   # No SMB, nothing here...
    #   return {}
    # }
    if {[string match *r* $args]} {
      set nonServers {}
    }
    frputs args
    set shares [UNCpoll [string match *a* $args]]
    foreach shar $shares {
      lassign $shar share drive
      set sarry($share) $drive
    }
    # set con [regsub -all {\\} [twapi::wnet_connected_resources] /]
    # foreach sh $con {
    #   lassign $sh letter sha
    #   set sarry($sha) $letter
    # }
    return [array get sarry]
  }

  proc setDN {mp share server target} {
    set drive $mp
    global config
    set cv {format "%s(%s)(%s)" $share $server $drive}
    if {$config(DisplayNameShare) != {}} {
      set cv $config(DisplayNameShare)
    }
    set displayName  [eval $cv]
    # This deletes a possibly old display name with nil drive
    set drive "*"
    set displayNamePat [eval $cv]
    set co [delDNpat $displayNamePat]
    frputs co displayNamePat
    set r [catch {addDN  [list $target $displayName]} er]
    frputs er mp share server target drive
    if {$r != 0 && [string index $er 0] != "-"} {
      PopWarn $er
    }
  }
  # Here we find the WSL shares and set them up to appear in the <volume>
  # listing where to look for local WSL files
  #
  # I addition, for each wsl found we generate a mapping to follow wsl side
  # links that point back to the MSW side of the house.
  #
  # I.e. files that look like: //wsl$/<name>/mnt/c/... -> c:/...
  # except that '/mnt/c' is just the default and does not cover mapping to
  # MSW side remote shares. So we do it all here.
  # This only works on wsl 1 shares so we don't recommend using it.
  # We also note that there may be more than one wsl system.

  # And now there are two versions 1 & 2
  #
  proc findWSLshares {} {
    variable wslNativeMap {} ; # list of wsl istances {name rootfs version}
    variable connected\
	[string tolower \
	     [regsub -all {\\} [::twapi::wnet_connected_resources -type file] {/}]]
    # Get WSL info from the registery
    set wslKey\
	HKEY_CURRENT_USER/SOFTWARE/Microsoft/Windows/CurrentVersion/Lxss
    if {[catch {registry keys [file native $wslKey]} wslEnts] != 0} {return}
    foreach wslEnt $wslEnts {
      # not all the 'Lxss' keys are wsl entries. This ignores the others
      if {[catch {registry get [file native $wslKey/$wslEnt] DistributionName}\
	       wslName] != 0} {
	continue
      }
      #set wslVers  [registry get  [file native $wslKey/$wslEnt] Version] 
      set wslBase  [registry get  [file native $wslKey/$wslEnt] BasePath]
      set wslBase  [regsub {^//\?/} [regsub -all {\\} $wslBase {/}] {}]
      set wslVers  [registry get  [file native $wslKey/$wslEnt] Flags]
      set wslVers  [expr {($wslVers >> 3) + 1}]

      if {[set conId [lsearch -exact -index 1 $connected //wsl$/$wslName]] == -1} {
	# wsl not connected, ignore it
	continue
      } else {
	set drvId [string toupper [lindex $connected "$conId 0"]]
      }
      setDN $drvId $wslName "wsl$" "//wsl$/$wslName"
      if {[lsearch -exact -index 0 $wslNativeMap "//wsl$/$wslName"] == -1} {
	lappend wslNativeMap [list "//wsl$/$wslName" $wslBase $wslVers $drvId]
      }
    }
      
    # Still here?? Ok, get the info
    # now that we know the wsl shares, lets put together a mapping for
    # wsl files that reference back to the non-wsl files.
    # This will be used by the link de-code code in 'getWsLink'
    # in 'winLnkFile.tcl'
    # I suppose it is possible that the /etc/wsl.conf is itself a link which we
    # will try to figure (it is not clear if the wsl set up allows this).
    # Here we compile the translations from //wsl$/../ files that link back
    #
    # We assume that /etc/wsl.config (if it exists) is not a link to MSW land.
    #
    variable wslMswMap {}
    
    foreach {wslId dontcare vers drvId} $wslNativeMap {
      # Set the default...
      set root {}
      set mountFsTab 0
      set enabled 0
      if {[catch {read [set fid [open $wslId/etc/wsl.conf r]]} wslInfo] == 0} {
	foreach line [split $wslInfo \n] {
	  set line [string trim $line]
	  if {[string index $line 0] == "\;"} {continue}
	  if {[string match -nocase "enabled*" $line]} {
	    set sl [split $line =]
	    set enabled [lindex $sl 1]
	  }
	  if {[string match -nocase "root*" $line]} {
	    set sl [split $line =]
	    set root [lindex $sl 1]
	  }
	  if {[string match -nocase "mountFsTab*" $line]} {
	    set sl [split $line =]
	    set mountFsTab [lindex $sl 1]
	  }
	}
	frputs fid wslId
	catch {close $fid}
	if {$root == {} || !$enabled } {
	  set mountFsTab 0
	  set root "/mnt"
	}
	foreach vol [file volum] {
	  if {[::twapi::get_drive_type $vol] != "fixed"} {continue}
	  frputs vol root 
	  lappend wslMswMap $wslId$root/[string tolower [string index $vol 0]]\
	      [string range $vol 0 1]
	}
	# That should take care of the wsl.config file. Did it turn on fstab?
	if {!$mountFsTab} {continue}
	set fstab $wslId/etc/fstab
	if {[file type fstab] == "link"} {
	  lassign [getWsLink $wslId/etc "fstab"] fstab cd
	}
	if {[catch {read [set fid [open $fstab r]]} wslInfo] == 0} {
	  foreach line [split $wslInfo \n] {
	    set line [string trim $line]
	    if {[string index $line 0] == "#"} {continue}
	    # The lines we are looking for lead with the following:
	    # <vol>.. DrvFs <wsl mount point> etc....
	    set pos [file volum]
	    if {[string toupper\
		     [regsub {\\} [string range $line 0 2] {/}]] in $pos} {
	      lappend wslMswMap $wslId[lindex $line 2]\
		  [file native [lindex $line 0]]
	    }
	  }
	}
	catch {close $fid}
      }
    }
  }
}
# This code gets called on reading of the config file so it
# must not redo things it has already done

# It is called with no parms. It looks at all the cifs:// entries in
# the password locker. These entries look like:

# cifs://user@SERVER/SHARE password flags
#
# if 'flags' contains a drive letter followed by : ([a-z]:) we attempt to
# connect the share to that drive letter. If 'flags' is 'nil' we ignore
# the entry (presumed to exist but not to be connected). In all other cases
# the connection is made, but no drive letter is assigned. These show up as
# UNC connections.
# We take the 'drive' as a suggestion. If possible we assign that letter
# to the share unless it is already in use. In which case we allow the 
# system to assign a letter and return it.

# A word on case. This is MSW, there is NO case. That said, one may want
# a particular case. The UNC find code seems to return uppercase for the
# server and lower for the path (the part after the server: //server/path
# (Likely because 'nbtstat' returns upper case for the server.)
# So here is how we allow the user to do what they want.
# When we check if this is a new share (in the passwordLocker) we use '-ncase'
# This allows the user to change it to what ever case s/he wants.
# (by modifying the passwordLocker).

# This is a resolve hostname twapi call that will use the async option if
# possible (i.e. if tcl is threaded).

proc getHostIp {hostname} {
  puts "here $hostname"
  frputs hostname
  set ::noAns 1
  if {1 || [catch {::twapi::resolve_hostname $hostname -async getHostResultx} xx] != 0} {
    puts "xx"
    return [::twapi::resolve_hostname $hostname]
  }
  puts "2 >$xx<"
  # Now we need to wait... we will time limit this to 30 sec.
  set afterId [after 3000 {set ::noAns {}}]
  while {$::noAns == 1} {realWaitForIdle}
  after cancel $afterId
  puts "$noAns"
  return $::noAns
}
proc getHostResult {h status ip} {
  frputs h status ip
  puts "$h $statue $ip"
  if {$status == "success"} {
    set ::noAns $ip
  }
  set ::noAns {}
}
  


proc setUpShare {} {
  global config findUNC::connected
  set mess {}
  # lappend mess "start setupshare"
  # startProc "Start setUpShare"

  # here we separate the drive and the flags
  foreach ci [::pwLocker::patternGet "cifs://*"] {
    lassign $ci unc pw arg
    set pw [lindex $pw 0]
    # startProc "Start work on $unc"
    # lappend mess "set up $ci connected>$connected<"
    unset -nocomplain done
    set flags {}
    set continue 0
    # Flag meanings (we only check for the first letter and we casefold):
    #  No        do not generate a Display Name.
    #  Only      generate a Display Name (It will appear in the volume list).
    #  Hot or {} generate a Display Name and put the display name in the HotList.
    #  Don't     do not ever set up this share
    #             (to change this use 'Utilies -> Show/edit Pass)

    # If we don't find a drive letter, we set up the UNC name (\\server\share)
    # a "*" or "*:" will let the system assign the drive letter
    set pwA(-localdevice) {}
    foreach ar [string tolower $arg] {
      if {[string match -nocase {[a-z]:} $ar] || $ar in {"*" "*:"}} {
	set pwA(-localdevice) [expr {$ar == "*:" ? "*" : $ar}]
	continue
      }
      set fl [string toupper [string index $ar 0]]
      if {"D" == $fl} {
	set continue 1
	break
      }
      if {$fl in {"N" "O" "H"}} {
	lappend flags $ar
      }
    }
    # if we did the above break, we are to forget about this share
    # lappend mess "$unc >$continue<"
    if {$continue} {continue}
   
    IsVFS $unc
    # build a clean share name
    set path [string tolower $VFSpath]
    lassign [split $VFSadd @] pwA(-user) VFSadd
    # If not path part, this is not a legal cifs entry, pass back the bad news.
    if {$path == {}} {
      lappend mess "Password Locker entry \"$unc\" is not a legal 'cifs' share name.\
                    \nEntry ignored."
      continue
    }
    if {$VFSadd == {}} {
      # no user...
      set VFSadd $pwA(-user)
      set pwA(-user) {}
    }
    set share //$VFSadd/$path
    set conAs [lsearch -all -exact -inline -nocase -index 1 $connected $share]
    set conLet {}
    set dr {}
    foreach con $conAs {
      set dd [lindex $con 0]
      # if connected with the requested drive we are done
      if {$pwA(-localdevice) == $dd} {
	incr done
	break
      }
      if {$dd != {}} {
	lappend dr $dd
      }
    }
    frputs dr "[llength $dr] " pwA(-localdevice) conAs done
    if {![info exist done]} {
       if {$pwA(-localdevice) ni [list {} {*}$dr] && $dr != {}} {
	# if share is connected but requested drive is not the one and there are
	# connected drives for this share, use one.
	set $pwA(-localdevice) [lindex $dr 0]
      }
      # if dr is nil and conAS is not, we have a connected share with no drive letter
      # if conAs is nil, the share is not connected.
      # in either case we attempt to connect with the requsted share.
      frputs VFSadd path VFSpath user unc reqDrive pwA(-localdevice) conAs dr
      # lappend mess "22 $share >$conAs< >$connected<"
      if {$conAs == {} || $dr == {}} {
	# not connected. Check to see if the requested system is up.
	if {$VFSadd ni {wsl$ WSL$}} {
	  # startProc "About to resolve hostname for $VFSadd"
	  if {[getHostIp $VFSadd] == {}} {
	    # host not found, log quietly and go to the next entry
	    Log "System '$VFSadd' not found, skipping set up of '$share'."
	    continue
	  }
	  # startProc "After hostname"
	  if {[catch {::twapi::get_shares -system $VFSadd -type file} shareThis] != 0} {
	    Log "Error '$shareThis' from '$VFSadd' attempting to get shares."
	    continue
	  }
	  # startProc "After get-shares $VFSadd"
	  if {$path ni [string tolower $shareThis]} {
	    Log "Share '$share' is not available on system '$VFSadd'\
      	      \n(only '$shareThis' available)."
	  }
	}
	set pwA(-type) file
	# We are not connected, but the host is up and sharing what we want
	# Lets try to connect.  Use a while loop incase we need a better password
	array unset locAry
	while {1} {
	  # The pwA array has the drive letter or "*" (-localdevice),
	  # the password (-password) and the user (-user)
	  set result 0
	  set pwA(-password) [::twapi::conceal $pw]
	  set ops [expr {$pw == -1 ? "-nopassword" : ""}]
	  # startProc "About to connect share $share"
	  set r [catch {::twapi::connect_share [file native $share] \
			    {*}[array get pwA] [list {*}$ops]} out]

	  # startProc "After connect share $share"
	  # lappend mess "connect share returned $out"
	  if {$r == 0} {
	    # success, on to the Display name
	    if {[info exist pwA(-localdevice)] && $pwA(-localdevice) == "*"} {
	      # if we asked the system to supply a drive letter, grab the result
	      set $pwA(-localdevice) $out
	    }
	    break
	  }
	  # Assume password failed, but put in message just in case
	  set r [catch {
	    ::pwLocker::getPw [expr {[info exists locAry(id)] ? $locAry(id) : $unc}] \
		user $pwA(-user) \
		nocase 1\
		newuser 1 \
		startwith [list $pw] \
		prompt $out \
		display [list "Drive" $pwA(-localdevice) "Flag" $flags]\
		returnall 1} pwRtn]
	  
	  if {$r} {
	    # need a break+continue here
	    incr result
	    lappend mess "Password request failed for share\
      	        '[expr {[info exists locAry(id)] ? $locAry(id) : $unc}]' with:\
         	 \"$pwRtn\" $::errorInfo"
	    break
	  }
	  if {$pwRtn == -2} {
	    incr result
	    lappend mess "User canceled set up of '$share'."
	    break
	  }
	  
	  if {[llength $pwRtn] > 3} {
	    array set locAry $pwRtn
	    set pw $locAry(password)
	    set pwA(-localdevice) $locAry(Drive)
	    lassign [URLsplit $locAry(id)] VFSpro pwA(-user) VFSadd p VFSpath
	    if {![string match -nocase $VFSpro "cifs"]} {
	      # Ouch! got back something totally new. Complain and bail
	      lappend mess "Password query for share '$share' passed back a '$VFSpro'
                         \naddress. Must be 'cifs'. Treating as user cancel."
	      incr result
	      break
	    }
	    set share "//$VFSadd/$VFSpath"
	    # if we are not to set up...
	    if {[string index $locAry(Flag) 0] in {"D" "d"}} {
	      set result 2
	      break
	    }
	  }

	} ; # End of while, if result is true, continue with the next one.
	# lappend mess "result >$result<"
	if {$result == 1} {continue}
	
	# Here we may need to update the password locker...
	if {[info exists locAry(id)] && $locAry(keep)} {
	  # We did call the and s/he wants to keep the result
	  # first get rid of what we started with
	  ::pwLocker::delete $unc
	  ::pwLocker::putPw $locAry(id) $pw [list $pwA(-localdevice) $locAry(Flag)]
	  set flags $locAry(Flag)
	}
	if {$result == 2} {continue}
      } ; # A new connection was made or not, now lets do the display name
    }
    # info for display name is in 'flags' and 'share'
    
    # Ok, the system is set, lets set up display names as needed
    # drive will be the $drive letter, and the share is $share
    # lappend mess "setting up $share"
    set fl [string toupper [string index $flags 0]]
    if {$fl == "N"} {continue}
    # reformat share for the format statement
    set target $share
    lassign [split $share /] > > server share
    # if {[info exists dr] && $dr != {}} {
    #   set pwA(-localdevice) $dr
    # }
    set drive {}
    if {[info exist pwA(-localdevice)]} {
      set drive [string toupper $pwA(-localdevice)]
      frputs drive pwA(-localdevice)
    }
    if {![string equal $drive {}]} {
      frputs drive target "[expr {{$drive} == {}}] "
      set target $drive
    }
    frputs drive target server share fl
    ::findUNC::setDN $drive $share $server $target
    if {(![info exists ::hotlistExclusion] || \
	     $target ni $::hotlistExclusion) && \
	    $fl == "O"} {
      lappend ::hotlistExclusion $target
    }
  }
  # startProc "End of setUpShare"
  return $mess
}
